/*
** Copyright (C) Red Hat, Inc. 2005-2008
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "clusterMonitorSnmp.h"
#include "clusterMIB.h"
#include "nodesMIB.h"
#include "servicesMIB.h"
#include "clumond_globals.h"
#include <syslog.h>
#include "Cluster.h"
#include "rhc_oid.h"
#include "Trap.h"
#include "nodesMIB.h"

#include <list>

using namespace ClusterMonitoring;
using namespace std;

ClusterMonitoring::ClusterMonitor monitor;

/* From nodesMIB_access.cpp */
String getNodeStatusDescription(unsigned int code);
unsigned int getNodeStatusCode(counting_auto_ptr<Node> node);

/* from servicesMIB_access.cpp */
String getServiceStatusDescription(unsigned int code);
unsigned int getServiceStatusCode(counting_auto_ptr<Service>);


void
check_cluster_state(unsigned int clientreg, void *clientarg)
{
	counting_auto_ptr<Cluster> old_status = monitor.get_cache();
	counting_auto_ptr<Cluster> cluster = monitor.get_cluster(0);

	if (!old_status.get() || !cluster.get())
		return;

	if (old_status->config_version() != cluster->config_version()) {
		// Send configuration version update trap
		Trap trap(rhcConfigVersion_trap, rhcConfigVersion_len);
				
		trap.Add(rhcClusterConfigVersion_oid,
			 rhcClusterConfigVersion_len,
			 cluster->config_version());
		trap.Send();
	}

	if (old_status->quorate(true) != cluster->quorate(true)) {
		// Send configuration version update trap
		Trap trap(rhcQuorumState_trap, rhcQuorumState_len);
				
		trap.Add(rhcClusterQuorate_oid,
			 rhcClusterQuorate_len,
			 cluster->quorate(true));
		trap.Send();
	}

	/* ****** NODES ****** */

	/* Compare nodes */
	list<counting_auto_ptr<Node> > old_nodes = old_status->nodes();
	list<counting_auto_ptr<Node> > nodes = cluster->nodes();
	int found = 0;

	/* Uses table/name; dunno if this is legit */
	oid rhcNodeStatus_oid[]	= { 1, 3, 6, 1, 4, 1, 2312, 8, 3, 1, 2 };
	oid rhcNodeID_oid[] =     { 1, 3, 6, 1, 4, 1, 2312, 8, 3, 1, 6 };

	for (list<counting_auto_ptr<Node> >::iterator x = old_nodes.begin();
	     x != old_nodes.end(); x++) {
		found = 0;

		for (list<counting_auto_ptr<Node> >::iterator y = nodes.begin();
		     y != nodes.end(); y++) {
			if ((*x)->name() != (*y)->name())
				continue;

			int statuscode = getNodeStatusCode((*y));

			found = 1;

			if ((*x)->online() == (*y)->online())
				continue;
			if ((*x)->clustered() == (*y)->clustered())
				continue;

			Trap trap(rhcNodeState_trap, rhcNodeState_len);

			trap.Add(rhcClusterNodesNames_oid,
				 OID_LENGTH(rhcClusterNodesNames_len),
				 (*x)->name());
			trap.Add(rhcNodeStatus_oid,
				 OID_LENGTH(rhcNodeStatus_oid),
				 statuscode);
			trap.Add(rhcNodeID_oid,
				 OID_LENGTH(rhcNodeID_oid),
				 (*x)->nodeid());
			trap.Send();
		}

		if (!found) {
			/* present in old but not new; advertise as offline */
			Trap trap(rhcNodeState_trap, rhcNodeState_len);
			trap.Add(rhcClusterNodesNames_oid,
				 OID_LENGTH(rhcClusterNodesNames_len),
				 (*x)->name());
			trap.Add(rhcNodeStatus_oid,
				 OID_LENGTH(rhcNodeStatus_oid), 2);
			trap.Add(rhcNodeID_oid,
				 OID_LENGTH(rhcNodeID_oid), (*x)->nodeid());
			trap.Send();
		}
	}

	for (list<counting_auto_ptr<Node> >::iterator x = nodes.begin();
	     x != nodes.end(); x++) {
		found = 0;

		for (list<counting_auto_ptr<Node> >::iterator y = old_nodes.begin();
		     y != old_nodes.end(); y++) {
			if ((*x)->name() != (*y)->name())
				continue;

			found = 1;
		}

		if (!found) {
			/* present in new but not old; update state if online */
			int statuscode = getNodeStatusCode((*x));

			if (statuscode == 2)
				continue;

			Trap trap(rhcNodeState_trap, rhcNodeState_len);
			trap.Add(rhcClusterNodesNames_oid,
				 OID_LENGTH(rhcClusterNodesNames_len),
				 (*x)->name());
			trap.Add(rhcNodeStatus_oid,
				 OID_LENGTH(rhcNodeStatus_oid), 2);
			trap.Add(rhcNodeID_oid,
				 OID_LENGTH(rhcNodeID_oid), (*x)->nodeid());
			trap.Send();
		}
	}

	/* ****** Services ****** */

	/* Compare services */
	list<counting_auto_ptr<Service> > old_svcs = old_status->services();
	list<counting_auto_ptr<Service> > svcs = cluster->services();

	/* Uses table/name; dunno if this is legit */
	oid rhcSvcStatus_oid[] = { 1, 3, 6, 1, 4, 1, 2312, 8, 3, 2, 3 };
	oid rhcSvcOwner_oid[] =  { 1, 3, 6, 1, 4, 1, 2312, 8, 3, 2, 5 };

	for (list<counting_auto_ptr<Service> >::iterator x = old_svcs.begin();
	     x != old_svcs.end(); x++) {
		found = 0;

		for (list<counting_auto_ptr<Service> >::iterator y = svcs.begin();
		     y != svcs.end(); y++) {

			int old_code, new_code;

			if ((*x)->name() != (*y)->name())
				continue;

			found = 1;

			old_code = getServiceStatusCode((*x));
			new_code = getServiceStatusCode((*y));

			if (old_code == new_code)
				continue;
			if ((*x)->nodename() == (*y)->nodename())
				continue;

			Trap trap(rhcServiceState_trap, rhcServiceState_len);

			trap.Add(rhcClusterServicesNames_oid,
				 OID_LENGTH(rhcClusterServicesNames_len),
				 (*y)->name());
			trap.Add(rhcSvcStatus_oid,
				 OID_LENGTH(rhcSvcStatus_oid),
				 getServiceStatusDescription(new_code));
			trap.Add(rhcSvcOwner_oid,
				 OID_LENGTH(rhcSvcOwner_oid),
				 (*y)->nodename());
			trap.Send();
		}

		if (!found) {
			/* present in old but not new; set status as unknown */
			Trap trap(rhcServiceState_trap, rhcServiceState_len);

			trap.Add(rhcClusterServicesNames_oid,
				 OID_LENGTH(rhcClusterServicesNames_len),
				 (*x)->name());
			trap.Add(rhcSvcStatus_oid,
				 OID_LENGTH(rhcSvcStatus_oid), 
				 getServiceStatusDescription(3));
			trap.Add(rhcSvcOwner_oid,
				 OID_LENGTH(rhcSvcOwner_oid), "none");
			trap.Send();
		}
	}

	for (list<counting_auto_ptr<Service> >::iterator x = svcs.begin();
	     x != svcs.end(); x++) {
		found = 0;

		for (list<counting_auto_ptr<Service> >::iterator y = old_svcs.begin();
		     y != old_svcs.end(); y++) {
			if ((*x)->name() != (*y)->name())
				continue;

			found = 1;
		}

		if (!found) {
			/* present in new but not old; update state if online */
			int code = getServiceStatusCode((*x));

			Trap trap(rhcServiceState_trap, rhcServiceState_len);
			trap.Add(rhcClusterServicesNames_oid,
				 OID_LENGTH(rhcClusterServicesNames_len),
				 (*x)->name());
			trap.Add(rhcSvcStatus_oid,
				 OID_LENGTH(rhcSvcStatus_oid),
				 getServiceStatusDescription(code));
			trap.Add(rhcSvcOwner_oid,
				 OID_LENGTH(rhcSvcOwner_oid), (*x)->nodename());
			trap.Send();
		}
	}
}


void
init_RedHatCluster(void)
{
	try {
		initialize_clusterMIB();
		initialize_nodesMIB();
		initialize_servicesMIB();

		// Cache information forever; let our trap alarm
		// do all updates.  This polling model is inadequate
		// and will have to be rewritten eventually.
		monitor.expire_cache((unsigned int)-1);

    		snmp_alarm_register(5, SA_REPEAT, check_cluster_state, NULL);

	} catch ( ... ) {}
}

void
init_redhatCluster(void) {
	init_RedHatCluster();
}
